Meta Entity
===========

Description
-----------

Meta Entity module allows to add metadata about an entity, stored in a
dedicated entity (`meta_entity`). This is useful when you want to avoid storing
this information as a content entity field and clutter the content entity with
data that is not part of the content but is [metadata
](https://en.wikipedia.org/wiki/Metadata) (data about other data).

Requirements
------------

* [Dynamic Entity
Reference](https://www.drupal.org/project/dynamic_entity_reference) module.

How to use it?
--------------

Even the module provides a user interface, it's mainly intended to be used
programmatically. Meta data is represented by entities, of type `meta_entity`
that are linking content entities. Meta entity types can be defined as bundle
configuration entities (of type `meta_entity_type`) in order to define types of
metadata (e.g. visits count, download count, etc). Each `meta_type` entity is
linked to a content entity (such as node, taxonomy term, etc) via a [Dynamic
Entity Reference](https://www.drupal.org/project/dynamic_entity_reference)
field. A content entity can be referred by a single meta entity of that type.
When defining a meta entity type, the site builder can choose the bundles of
content entities that can be linked by that specific type of metadata.

Reverse references
------------------

On each bundle selected for a given meta entity bundle, the site builder may
configure a field name. The target entity will automatically expose an
`entity_reference` computed field with that name, linking back to the meta
entity.

API usage
---------

Create a meta entity type to be used on 'article' nodes:

```
MetaEntityType::create([
  'id' => 'visits',
  'label' => 'Visits',
  'mapping' => [
    // Node entity type ID.
    'node' => [
      // The array is keyed by bundle ID. A field identifier may be configured
      // as value. For instance, the 'article' nodes will automatically expose a
      // field, named 'visits' which is a reference to the node's meta entity.
      'article' => 'visits',
      // If the value is NULL, no computed field is created but it's still
      // allowed to create meta entities for 'page' nodes.
      'page' => NULL,
    ],
    // Entity types without bundles, such as 'user', should use the entity type
    // ID as bundle ID.
    'user' => [
      'user' => 'field_mentions',
    ],
  ],
])->save();
```

Add all needed fields to this meta entity type, either as config entities or
using the UI. For instance, in the above example you could add a configurable
field `field_count` of type `integer`.

Create an 'article' node.
```
$node = Node::create([
  'type' => 'article',
  'title' => 'News From The World',
  ...
]);
$node->save();
```

Create and link the meta entity:

```
$visits = MetaEntity::create([
  'type' => 'visits',
  'target' => $node,
  'field_count' => 10,
]);
$visits->save();
```

Now the node `$node` has metadata, represented by a meta entity of type
`visits`, associated without the node to be touched. The meta entity can update
(e.g. increment the `field_count`) without needing to re-save the `$node`. On
meta entity update, also the `$node` cache tags are invalidated.

The content entity can refer directly the meta entity, via the computed reverse
reference:

```
$message = t('@count visits', [
  '@count' => $node->visits->entity->field_count->value;
]);
```

For this reason, creating an entity with its meta entity could be achieved in a
single step:

```
Node::create([
  'type' => 'article',
  'title' => 'News From The World',
  'visits' => MetaEntity::create([
    'type' => 'visits',
    'field_count' => 10,
  ]),
])->save();
```

Supposing a new meta entity type, `downloads`, is added and a new metadata
entity of this type is linked to `$node`:

```
MetaEntityType::create([
  'id' => 'downloads',
  'label' => 'Downloads',
  'mapping' => [
    'node' => [
      'article' => 'download_count',
    ],
  ],
])->save();

$downloads = MetaEntity::create([
  'type' => 'downloads',
  'target' => $node,
  'field_count' => 100,
]);
$downloads->save();
```

Now there are two reverse reference fields exposed on the node:

```
// Will return $downloads.
$node->download_count->entity;
// Will return $visits.
$node->visits->entity;
```
