<?php

declare(strict_types = 1);

namespace Drupal\meta_entity;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\meta_entity\Entity\MetaEntityInterface;

/**
 * Provides an interface for 'meta_entity.repository' service.
 */
interface MetaEntityRepositoryInterface {

  /**
   * Returns a list of meta entities of a given entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to checked.
   * @param string $meta_entity_type_id
   *   The meta entity type ID.
   *
   * @return \Drupal\meta_entity\Entity\MetaEntityInterface|null
   *   The meta entity or NULL if it doesn't exist.
   */
  public function getMetaEntityForEntity(ContentEntityInterface $entity, string $meta_entity_type_id): ?MetaEntityInterface;

  /**
   * Returns a list of existing meta entities of a given entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to checked.
   *
   * @return \Drupal\meta_entity\Entity\MetaEntityInterface[]
   *   Associative array keyed by meta entity ID with meta entity as value.
   */
  public function getMetaEntitiesForEntity(ContentEntityInterface $entity): array;

  /**
   * Returns a list of meta entity types available for a given entity bundle.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string|null $bundle
   *   The entity bundle. Can be omitted when an entity type without bundles,
   *   such as 'user', is passed.
   *
   * @return string[]
   *   A list of meta entity type IDs.
   */
  public function getMetaEntityTypesForBundle(string $entity_type_id, ?string $bundle = NULL): array;

  /**
   * Returns a list of reverse reference field names for a given entity bundle.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string $bundle
   *   The entity bundle.
   *
   * @return string[]
   *   An associative array, keyed by bundle ID, of field names.
   */
  public function getReverseReferenceFieldNames(string $entity_type_id, string $bundle): array;

  /**
   * Returns a list of meta entity types that are configured with auto-creation.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity for which to return the list.
   *
   * @return string[]
   *   A list of meta entity type IDs.
   */
  public function getTypesWithAutoCreation(ContentEntityInterface $entity): array;

}
