<?php

declare(strict_types = 1);

namespace Drupal\meta_entity\Plugin\Validation\Constraint;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\meta_entity\MetaEntityRepositoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;

/**
 * Validates the target entity type and bundle.
 *
 * The target entity type and bundle should match the ones that are defined in
 * the mapping for this meta entity type.
 */
class MappedTargetEntityValidator extends ConstraintValidator implements ContainerInjectionInterface {

  /**
   * The meta entity repository service.
   *
   * @var \Drupal\meta_entity\MetaEntityRepositoryInterface
   */
  protected $metaEntityRepository;

  /**
   * Constructs a new validator instance.
   *
   * @param \Drupal\meta_entity\MetaEntityRepositoryInterface $meta_entity_repository
   *   The meta entity repository service.
   */
  public function __construct(MetaEntityRepositoryInterface $meta_entity_repository) {
    $this->metaEntityRepository = $meta_entity_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static($container->get('meta_entity.repository'));
  }

  /**
   * {@inheritdoc}
   */
  public function validate($items, Constraint $constraint): void {
    /** @var \Drupal\Core\Entity\ContentEntityInterface $target_entity */
    $target_entity = $items->entity;
    if (empty($target_entity)) {
      return;
    }
    $entity_types = $this->metaEntityRepository->getMetaEntityTypesForBundle($target_entity->getEntityTypeId(), $target_entity->bundle());
    if (!in_array($items->getEntity()->bundle(), $entity_types)) {
      $this->context->buildViolation($constraint->message, [
        '%type' => $target_entity->getEntityType()->getLabel(),
        '%bundle' => $target_entity->bundle(),
        '%meta' => $items->getEntity()->get('type')->entity->label(),
      ])->setCode('MappedTargetEntity')
        ->addViolation();
    }
  }

}
